
from enum import Enum
import logging
import re

from pydantic import Field
from labfreed.labfreed_infrastructure import LabFREED_BaseModel, ValidationMessage, ValidationMsgLevel
from labfreed.pac_id.pac_id import PAC_ID
from labfreed.pac_id_resolver.services import Service, ServiceGroup

class ServiceType(Enum):
    USER_HANDOVER_GENERIC = 'userhandover-generic'
    ATTRIBUTE_SERVICE_GENERIC = 'attributes-generic'


class CITEntry_v1(LabFREED_BaseModel):
    applicable_if: str = Field(..., min_length=1)
    service_name: str = Field(..., min_length=1)
    application_intent:str = Field(..., min_length=1)
    service_type:ServiceType
    template_url:str = Field(..., min_length=1)


class CIT_v1(LabFREED_BaseModel):
    origin:str = ''
    entries:list[CITEntry_v1]
    
    
    @classmethod
    def from_csv(cls, csv:str, origin=''):
        lines = csv.splitlines()
        entries = list()
        errors = list()
        for line in lines:
            if not line: # empty line
                continue
            if line.strip()[0] == '#': #comment line
                continue
            if 'Service Name' in line.strip() : #header line
                continue
            
            cols = [c.strip() for c in line.split('\t')]
            try:
                entry = CITEntry_v1(
                                    service_name         = cols[0],
                                    application_intent   = cols[1],
                                    service_type         = ServiceType(cols[2]),
                                    applicable_if        = cols[3],
                                    template_url         = cols[4]
                                    )
            except ValueError:
                logging.error(f'invalid line {line}')
                msg = ValidationMessage(
                    level=ValidationMsgLevel.WARNING,
                    source='CIT line',
                    source_id=0,
                    msg='Invalid line in CIT. Line was ignored. Remaining CIT is functional.',
                    highlight_sub_patterns=line
                )
                errors.append(msg)

            entries.append(entry)           
        cit = CIT_v1(origin=origin, entries=entries)
        cit._validation_messages.extend(errors)
        cit._csv_original = csv
        return cit
    
    def evaluate_pac_id(self, pac):
        cit_evaluated = ServiceGroup(origin=self.origin)   
        for e in self.entries:
            conditions = e.applicable_if.split(';')
            conditions_evaluated = list()
            for c in conditions:
                if '=' in c:
                    query, expected = c.split('=')
                    value = self._find_in_pac(query.strip(), pac)
                    conditions_evaluated.append(value == expected.strip())
                else:
                    query = c.strip()
                    found = self._find_in_pac(query, pac)
                    conditions_evaluated.append(found)
            is_applicable = all(conditions_evaluated)
                    
            if not is_applicable:
                continue

            url = re.sub(r"\{([^}]+)\}", lambda v: self._find_in_pac(v.group(0), pac), e.template_url)
            cit_evaluated.services.append(Service(  
                                                    service_name=e.service_name,
                                                    application_intents= [ e.application_intent ],
                                                    service_type=e.service_type,
                                                    url = url
                                )
                            )            
        return cit_evaluated
    
    def _find_in_pac(self, value, pac:PAC_ID):
        pac_url =pac.to_url()
        if value == '{isu}':
            return pac.issuer
        
        elif value == '{pac}':
            return pac_url.split('*')[0]
        
        elif value == '{id}':
            m = re.match(r'^HTTPS://.+?/(.+?)(\*.*)*$', pac_url)
            return m.group(1) if m else None
        
        elif m := re.match(r'\{idSeg(\d+)\}', value):
            i = int(m.group(1)) - 1 # CIT is 1 based
            seg = pac.identifier[i] if i < len(pac.identifier) else None
            if seg:
                return f"{(seg.key + ':') if seg.key else ''}{seg.value}"
            
        elif m := re.match(r'\{idVal(\w+)\}', value):
            k = m.group(1)
            seg = [s for s in pac.identifier if s.key and s.key == k]
            if seg:
                seg = seg[0]
                return seg.value   
            else:
                return None 
            
        elif value == '{ext}':
            m = re.match(r'^.*?(\*.*)*$', pac_url)
            ext_str = m.group(1) if m else None
            return m.group(1)[1:] if ext_str else None
        
        elif m := re.match(r'\{ext(\d+)\}', value):
            i = int(m.group(1)) - 1 # CIT is 1 based
            extensions = pac_url.split('*') 
            extensions.pop(0)# first element is not extension
            return extensions[i] if i < len(extensions) else None
        else:
            return None
                

            
        
    
    def __str__(self):
        if csv:=self._csv_original:
            return csv
        
        s = "# coupling information table version: 1.0\n"
        s += "Service Name\tApplication Intent\tService Type\tApplicable If\tTemplate Url\n"
        for e in self.entries:
            s += '\t'.join([e.service_name, e.application_intent, e.service_type.value, e.applicable_if, e.template_url]) + '\n'
        return s




from enum import auto, Enum

from pydantic import Field
from requests import RequestException, get, head

from concurrent.futures import ThreadPoolExecutor, as_completed

from rich import print
from rich.table import Table

from labfreed.labfreed_infrastructure import LabFREED_BaseModel


class ServiceStatus(Enum):
    ACTIVE = auto()
    INACTIVE = auto()
    UNKNOWN = auto()
    
class Service(LabFREED_BaseModel):
    service_name: str
    application_intents:list[str]
    service_type:str
    url:str
    status:ServiceStatus =ServiceStatus.UNKNOWN
    
    def check_service_status(self):
        '''Checks the availability of the service.'''
        try:
            r = head(self.url, timeout=2)
            if r.status_code < 400:
                self.status = ServiceStatus.ACTIVE
            else: 
                self.status = ServiceStatus.INACTIVE
        except RequestException as e:
            print(f"Request failed: {e}")
            self.status = ServiceStatus.INACTIVE
             
    
class ServiceGroup(LabFREED_BaseModel):
    """ Services with common origin. The result of resolving a PAC-ID against a CIT"""
    origin: str = ""
    services: list[Service] = Field(default_factory=list)
    
    def update_states(self):
        '''Triggers each service to check if the url can be reached'''
        if not _has_internet_connection():
            raise ConnectionError("No Internet Connection")
        with ThreadPoolExecutor(max_workers=10) as executor:
            futures = [executor.submit(s.check_service_status) for s in self.services]
            for _ in as_completed(futures):
                pass  # just wait for all to finish
            

    
    def __str__(self):
        out = [f'CIT (origin {self.origin})']
        for s in self.services:
            out.append(f'{s.service_name}\t\t\t{s.url}')
        return '\n'.join(out)
    
    def print(self):
        table = Table(title=f"Services from origin '{self.origin}")

        table.add_column("Service Name")
        table.add_column("URL")
        table.add_column('Reachable')
        
        for s in self.services:
            table.add_row(s.service_name, s.url, s.status.name)

        print(table)
        
        
def _has_internet_connection():
    try:
        get("https://1.1.1.1", timeout=3)
        return True
    except RequestException:
        return False


from enum import Enum
import json
import re
from typing import Self
from pydantic import Field, field_validator, model_validator
import yaml
import jsonpath_ng.ext as jsonpath


from labfreed.pac_id_resolver.services import Service, ServiceGroup
from labfreed.labfreed_infrastructure import LabFREED_BaseModel, ValidationMsgLevel, _quote_texts

__all__ = [
    "CIT_v2",
    "CITBlock_v2",
    "CITEntry_v2"
]

class ServiceType(Enum):
    USER_HANDOVER_GENERIC = 'userhandover-generic'
    ATTRIBUTE_SERVICE_GENERIC = 'attributes-generic'


class CITEntry_v2(LabFREED_BaseModel):
    service_name: str
    application_intents:list[str]
    service_type:ServiceType |str
    template_url:str
    
    @model_validator(mode='after')
    def _validate_service_name(self):
        # service_name
        if not_allowed_chars := set(re.sub(r'[A-Za-z0-9\-\x20]', '', self.service_name)):
            self._add_validation_message(
                level=ValidationMsgLevel.ERROR,
                source=f'Service {self.service_name}',
                msg=f'Service name ontains invalid characters {_quote_texts(not_allowed_chars)}',
                highlight_sub=not_allowed_chars
            )
        
        if len(self.service_name) == 0 or len(self.service_name) > 255:
             self._add_validation_message(
                level=ValidationMsgLevel.ERROR,
                source=f'Service {self.service_name}',
                msg='Service name must be at least one and maximum 255 characters long'
                )
        return self
        
    
    @model_validator(mode='after')
    def _validate_application_intent(self):
        for intent in self.application_intents:
            if re.fullmatch('.*-generic$',  intent):
                self._add_validation_message(
                level=ValidationMsgLevel.ERROR,
                source=f'Application intent {intent}',
                msg="Ends with '-generic'. This is not permitted, since it is reserved for future uses'",
                highlight_sub=[intent]
                )

            if not_allowed_chars := set(re.sub(r'[A-Za-z0-9\-]', '', intent)):
                self._add_validation_message(
                    level=ValidationMsgLevel.ERROR,
                    source=f'Application intent {self.service_name}',
                    msg=f'Contains invalid characters {_quote_texts(not_allowed_chars)}',
                    highlight_sub=not_allowed_chars
                )
                
            if len(intent) == 0 or len(intent) > 255:
                self._add_validation_message(
                    level=ValidationMsgLevel.ERROR,
                    source=f'Application intent  {intent}',
                    msg='Must be at least one and maximum 255 characters long'
                    )
        return self
    
    @model_validator(mode='after')
    def _validate_service_type(self):
        allowed_types = [ServiceType.ATTRIBUTE_SERVICE_GENERIC.value, ServiceType.USER_HANDOVER_GENERIC.value]
        if self.service_type not in allowed_types:
            if isinstance(self.service_type, ServiceType):
                s= self.service_type.value
            else:
                s= self.service_type
            for at in allowed_types:
                s = s.replace(at,'')
            self._add_validation_message(
                level=ValidationMsgLevel.ERROR,
                source=f'Service Type  {self.service_type}',
                msg=f'Invalid service type. Must be {_quote_texts(allowed_types)} must be at least one and maximum 255 characters long',
                highlight_sub=s
                )
        return self
    
    

class CITBlock_v2(LabFREED_BaseModel):
    applicable_if: str  = Field(default='True', alias='if')
    entries: list[CITEntry_v2]
    
    @field_validator('applicable_if', mode='before')
    @classmethod
    def _convert_if(cls, v):
        return v if v is not None else 'True'
    
    


class CIT_v2(LabFREED_BaseModel):
    '''Coupling Information Table (CIT)'''
    origin: str = ''
    model_config = {
        "extra": "allow"
    }
    '''@private'''
    cit: list[CITBlock_v2] = Field(default_factory=list)
    
    @model_validator(mode='after')
    def _validate_origin(self):
        if len(self.origin) == 0:
            self._add_validation_message(level=ValidationMsgLevel.WARNING,
                                        source='CIT origin',
                                        msg='Origin should not be empty'
                                        )
        return self
    
    
    @classmethod
    def from_yaml(cls, yml:str) -> Self:
        return cls.model_validate(yml)
    
    def __str__(self):
        yml = yaml.dump(self.model_dump()                        )
        return yml
    
    def evaluate_pac_id(self, pac):
        pac_id_json = pac.to_dict()
        cit_evaluated = ServiceGroup(origin=self.origin)   
        for block in self.cit:
            _, is_applicable = self._evaluate_applicable_if(pac_id_json, block.applicable_if)
            if not is_applicable:
                continue

            for e in block.entries:
                url = self._eval_url_template(pac_id_json, e.template_url)
                cit_evaluated.services.append(Service(  
                                                        service_name=e.service_name,
                                                        application_intents=e.application_intents,
                                                        service_type=e.service_type,
                                                        url = url
                                    )
                              )            
        return cit_evaluated
    
    
    def _evaluate_applicable_if(self, pac_id_json:str, expression) -> tuple[str, bool]:
        expression = self._apply_convenience_substitutions(expression)
        
        tokens = self._tokenize_jsonpath_expression(expression)
        expression_for_eval = self._expression_from_tokens(pac_id_json, tokens)
        applicable = eval(expression_for_eval, {}, {})
        
        return expression_for_eval, applicable
    
    
    def _apply_convenience_substitutions(self, query):
        ''' applies a few substitutions, which enable abbreviated syntax.'''
        
        # allow access to array elements by key
        q_mod = re.sub(r'\[(".+?")\]', r'[?(@.key == \1)]', query )
        
        # allow shorter path
        # substitutions = [
        #     (r'(?<=^)id', 'pac.id'),
        #     (r'(?<=^)cat', 'pac.id.cat'),
        #     (r'(?<=\.)id(?=\.)', 'identifier'),
        #     (r'(?<=\.)cat$', 'categories'),
        #     (r'(?<=\.)cat(?=\[)', 'categories'),
        #     (r'(?<=\.)seg$', 'segments'),
        #     (r'(?<=\.)seg(?=\[)', 'segments'),
        #     (r'(?<=^)isu', 'pac.isu'),
        #     (r'(?<=\.)isu', 'issuer'),
        #     (r'(?<=^)ext', 'pac.ext'),
        #     (r'(?<=\.)ext(?=$)', 'extensions'),
        #     (r'(?<=\.)ext(?=\[)', 'extensions'),
        # ]
        # for sub in substitutions:
        #     q_mod = re.sub(sub[0], sub[1], q_mod)
        
        return q_mod
    

    def _tokenize_jsonpath_expression(self, expr: str):
        token_pattern = re.compile(
            r"""
            (?P<LPAREN>\() |
            (?P<RPAREN>\)) |
            (?P<LOGIC>\bAND\b|\bOR\b|\bNOT\b) |
            (?P<OPERATOR>==|!=|<=|>=|<|>) |
            (?P<JSONPATH>
                \$                               # starts with $
                (?:
                    [^\s\[\]()]+                # path segments, dots, etc.
                    |
                    \[                           # open bracket
                        (?:                     # non-capturing group
                            [^\[\]]+            # anything but brackets
                            |
                            \[[^\[\]]*\]        # nested brackets (1 level)
                        )*
                    \]
                )+                              # one or more bracket/segment blocks
            ) |
            (?P<LITERAL>
                -?[\w\.\-]+   # domain-like literals
            )
            """,
            re.VERBOSE
        )

        tokens = []
        pos = 0
        while pos < len(expr):
            match = token_pattern.match(expr, pos)
            if match:
                group_type = match.lastgroup
                value = match.group().strip()
                tokens.append((value, group_type))
                pos = match.end()
            elif expr[pos].isspace():
                pos += 1  # skip whitespace
            else:
                raise SyntaxError(f"Unexpected character at position {pos}: {expr[pos]}")

        return tokens
    
    
    def _expression_from_tokens(self, pac_id_json:str, tokens: tuple[str, str]):
        out  = []
        for i in range(len(tokens)):
            prev_token = tokens[i-1] if i > 0 else (None, None)
            curr_token = tokens[i]
            next_token = tokens[i+1] if i < len(tokens)-1 else (None, None)
            if curr_token[1] == 'JSONPATH':
                res = self._evaluate_jsonpath(pac_id_json, curr_token[0])
                
                if prev_token[1] == 'OPERATOR' or next_token[1] == 'OPERATOR':
                    # if token is part of comparison return the value of the node
                    if len(res) == 0:
                        out.append('""')
                    else:
                        out.append(f'"{res[0].upper()}"')
                else:
                    # if token is not part of comparison evaluate to boolean
                    if len(res) == 0:
                        out.append(False)
                    else:
                        out.append(True)
                        
            elif curr_token[1] == 'LOGIC':
                out.append(curr_token[0].lower())  
                
            elif curr_token[1] == 'LITERAL':
                t = curr_token[0]
                if t[0] != '"':
                    t = '"' + t
                if t[-1] != '"':
                    t = t + '"' 
                out.append(t.upper())    
            else:
                out.append(curr_token[0])
                
        s = ' '.join([str(e) for e in out])
        return s
                    

    
    
    def _eval_url_template(self, pac_id_json, url_template):
        url = url_template
        placeholders = re.findall(r'\{(.+?)\}', url_template)
        for placeholder in placeholders:
            expanded_placeholder = self._apply_convenience_substitutions(placeholder)
            res = self._evaluate_jsonpath(pac_id_json, expanded_placeholder) or ['']
            url = url.replace(f'{{{placeholder}}}', str(res[0]))
            # res = self.substitute_jsonpath_expressions(expanded_placeholder, Patterns.jsonpath.value, as_bool=False)
            # url = url.replace(f'{{{placeholder}}}', res)
        return url
    

    
    def _evaluate_jsonpath(self, pac_id_json, jp_query):
        if isinstance(pac_id_json, str):
            pac_id_json = json.loads(pac_id_json)
        jsonpath_expr = jsonpath.parse(jp_query)
        matches = [match.value for match in jsonpath_expr.find(pac_id_json)]
        return matches
    
    

    
    
    




from abc import ABC, abstractproperty

from pydantic import model_validator

from labfreed.labfreed_infrastructure import LabFREED_BaseModel

    
class ExtensionBase(ABC):
    name: str
    type: str


    @abstractproperty
    def data(self) -> str:
        raise NotImplementedError("Subclasses must implement 'data'")
    
    def __str__(self):
        return f'{self.name}${self.type}/{self.data}'
    


class Extension(LabFREED_BaseModel,ExtensionBase):
    '''Implementation of Extension for unknown extension types'''
    name:str
    type:str
    data_:str
    
    @property
    def data(self) -> str:
        return self.data_
       
    @staticmethod
    def create(*, name, type, data):
        return Extension(name=name, type=type, data=data)
    
    @model_validator(mode='before')
    @classmethod
    def move_data_field(cls, values):
        if "data" in values:
            values["data_"] = values.pop("data")
        return values
    
    model_config = {
        "extra": "allow",  # Allow extra keys during pre-validation
    }
    




from .pac_id import PAC_ID
from .id_segment import IDSegment
from .extension import Extension

'''@private
From a SW engineering perspective it would be best to have no dependencies from other modules to pac_id.
However from a Python users convenience perspective it is better to have one place where a pac url can be parsed and magically the extensions are in a meaningful type (e.g. TREX in TREX aware format) and categories are known of possible.

>> We have given priority to convenient usage and therefore chose to have dependencies from pac_id to pac_cat and well_known_extensions
'''


__all__ = [
    "PAC_ID",
    "IDSegment",
    "Extension"
]






from datetime import date, datetime, time
import logging
import re

from pydantic import RootModel
from labfreed.well_known_keys.unece.unece_units import unece_unit
from labfreed.trex.python_convenience.data_table import DataTable
from labfreed.utilities.base36 import from_base36, base36, to_base36

from labfreed.trex.python_convenience.quantity import Quantity, unece_unit_code_from_quantity
from labfreed.trex.table_segment import ColumnHeader, TableSegment
from labfreed.trex.trex import TREX
from labfreed.trex.trex_base_models import AlphanumericValue, BinaryValue, BoolValue, DateValue, ErrorValue, NumericValue, TextValue
from labfreed.trex.value_segments import BoolSegment, ErrorSegment, TextSegment, NumericSegment, AlphanumericSegment, DateSegment, ValueSegment


class pyTREX(RootModel[dict[str, Quantity | datetime | time | date | bool | str | base36 | DataTable]]):
    ''' A wrapper around dict, which knows how to convert to and from TREX. 
        It restricts the types allowed as values. Keys must be str.
    '''
    model_config = {'arbitrary_types_allowed':True} # needed to allow Quantity and DataTable w/o implementing the pydantic schema
    '''@private'''

    
    @classmethod
    def from_trex(cls, trex:TREX):
        '''Creates a pyTREX from a TREX'''
        return {seg.key: _trex_segment_to_python_type(seg) for seg in trex.segments}
             
             
    def to_trex(self):
        '''Creates a TREX'''
        segments = list()
        for k, v in self.root.items():
            if v is None:
                value = _error_value_from_python_type(v)
                segments.append(ErrorSegment(key=k, value=value.value))
            elif isinstance(v, bool):
                value = _bool_value_from_python_type(v)
                segments.append(BoolSegment(key=k, value=value.value))
            elif isinstance(v, Quantity):
                unece_code = unece_unit_code_from_quantity(v)
                value = _numeric_value_from_python_type(v.value)
                segments.append(NumericSegment(key=k, value=value.value, type=unece_code))
            elif isinstance(v, (int, float)):
                value = _numeric_value_from_python_type(v)
                segments.append(NumericSegment(key=k, value=value.value, type='C63'))  # unitless
            elif isinstance(v, (datetime, time, date)):
                value = _date_value_from_python_type(v)
                segments.append(DateSegment(key=k, value=value.value))
            elif isinstance(v, str):
                if re.fullmatch(r'[A-Z0-9\-\.]*', v):
                    value = _alphanumeric_value_from_python_type(v)
                    segments.append(AlphanumericSegment(key=k, value=value.value))
                else:
                    v = to_base36(v)
                    value = _text_value_from_python_type(v)
                    segments.append(TextSegment(key=k, value=value.value))
            elif isinstance(v, base36):
                value = _text_value_from_python_type(v)
                segments.append(TextSegment(key=k, value=value.value))
                
            elif isinstance(v, DataTable):
                v:DataTable = v
                headers = list()
                for nm, rt in zip(v.col_names, v.row_template):
                    if isinstance(rt, bool): # must come first otherwise int matches the bool
                        t = 'T.B'
                    elif isinstance(rt, Quantity):
                        unece_code = unece_unit_code_from_quantity(rt)
                        t = unece_code
                    elif isinstance(rt, (datetime, time, date)):
                        t = 'T.D'     
                    elif isinstance(rt, str):
                        if re.fullmatch(r'[A-Z0-9\-\.]*', rt):
                            t = 'T.A'
                        else:
                            v = to_base36(rt)
                            t = 'T.X'
                    elif isinstance(rt, base36):
                        t = 'T.X'
                    
                    headers.append(ColumnHeader(key=nm, type=t))
                data = []
                for row in v.data:
                    r = []
                    for e in row:
                        if e is None:
                            r.append(_error_value_from_python_type(e))
                        elif isinstance(e, bool): # must come first otherwise int matches the bool
                            r.append(_bool_value_from_python_type(e))
                        elif isinstance(e, Quantity):
                            r.append(_numeric_value_from_python_type(e.value))
                        elif isinstance(e, (int, float)):
                            r.append(_numeric_value_from_python_type(e))
                        elif isinstance(e, (datetime, time, date)):
                            r.append(_date_value_from_python_type(e))
                        elif isinstance(e, str):
                            if re.fullmatch(r'[A-Z0-9\-\.]*', e):
                                r.append(_alphanumeric_value_from_python_type(e))
                            else:
                                e = to_base36(e)
                                r.append(_text_value_from_python_type(e))
                        elif isinstance(e, base36):
                            r.append(_text_value_from_python_type(e))
                    data.append(r)
                segments.append(TableSegment(key=k, column_headers=headers, data=data))
        return TREX(segments=segments)
    
    # make the usual dict methods available, for convenience
    def __getitem__(self, key): return self.root[key]
    def __setitem__(self, key, value): self.root[key] = value
    def update(self, *args, **kwargs):
        return self.root.update(*args, **kwargs)
    def keys(self): return self.root.keys()
    def values(self): return self.root.values()
    def items(self): return self.root.items()
    def __contains__(self, key): return key in self.root
    def __iter__(self): return iter(self.root)
    def __len__(self): return len(self.root)
    
  
  
# Helper functions to convert python types to TREX types    

def _numeric_value_from_python_type(v:int|float):
    return NumericValue(value = str(v))


def _date_value_from_python_type(v:date|time|datetime):    
    sd = ""
    st = ""
    if isinstance(v, date) or isinstance(v, datetime):
        sd = v.strftime('%Y%m%d')
    if isinstance(v, time) or isinstance(v, datetime):
        if v.microsecond:
            st = v.strftime("T%H%M%S.") + f"{v.microsecond // 1000:03d}"
        elif v.second:
            st = v.strftime("T%H%M%S")
        else:
            st = v.strftime("T%H%M")
                        
    return DateValue(value = sd + st)
    
    
def _bool_value_from_python_type(v:bool):
    return BoolValue(value = 'T' if v else 'F')


def _alphanumeric_value_from_python_type(v:str):
    return AlphanumericValue(value = v)


def _text_value_from_python_type(v:base36|str):
    if isinstance(v, str):
        logging.info('Got str for text value > converting to base36')
        out =  to_base36(v).root
    else:
        out =  v.root
    return TextValue(value = out)
        
        
def _binary_value_from_python_type(v:base36|str):
    if isinstance(v, str):
        out = v
    else:
        out = v.root
    return BinaryValue(value = out)
    

def _error_value_from_python_type(v:str):
    if v is None:
        v = '-'
    return ErrorValue(value = v)
    


# Helper functions to convert from TREX types to python types
def _trex_segment_to_python_type(v):
    '''Converts a TREX segment to a python value. Note the segment key must be handles outside.'''
    if isinstance(v, NumericSegment):
        num_val = _trex_value_to_python_type(v)
        u = unece_unit(v.type)
        unit = u.get('symbol')
        return Quantity(value=num_val, unit=unit)
    
    # value segments are derived from their respective value type
    elif isinstance(v, ValueSegment):
        return _trex_value_to_python_type(v)

    elif isinstance(v, TableSegment):
        table = DataTable(col_names=[ch.key for ch in v.column_headers])
        for row in v.data:
            r = []
            for e, h in zip(row, v.column_headers):
                if isinstance(e, NumericValue):
                    u = unece_unit(h.type)
                    unit = u.get('symbol')
                    r.append(Quantity(value=e.value, unit=unit))
                else:
                    r.append(_trex_value_to_python_type(e))
            table.append(r)
        return table
        


def _trex_value_to_python_type(v):
    '''Converts a TREX value to the corresponding python type'''
    if isinstance(v, NumericValue):
        if '.' not in v.value and 'E' not in v.value: 
            return int(v)
        else:
            return float(v.value)  
        
    elif isinstance(v,DateValue):
        d = v._date_time_dict
        if d.get('year') and d.get('hour'): # input is only a time
            return datetime(**d)
        elif d.get('year'):
            return date(**d)
        else:
            return time(**d)
            
    elif isinstance(v, BoolValue):
        if v.value == 'T':
            return True
        elif v.value == 'F':
            return False
        else:
            Exception(f'{v} is not valid boolean. That really should not have been possible -- Contact the maintainers of the library')
                
    elif isinstance(v, AlphanumericValue):
        return v.value
        
    elif isinstance(v, TextValue):
        decoded = from_base36(v.value)
        return decoded
        
    elif isinstance(v, BinaryValue):
        decoded = bytes(from_base36(v.value))
        return decoded
        
    elif isinstance(v, ErrorValue):
        return v.value
        
    else:
        raise (TypeError(f'Invalid type {type(v)} of segment'))
 


from enum import Enum


class WellKnownKeys(Enum):
    GTIN = '01'
    BATCH = '10'
    SERIAL = '21'
    ADDITIONAL_IDINTIFIER = '240'
    RUN_ID_ABSOLUTE = 'RNR'
    SAMPLE_ID = 'SMP'
    EXPERIMENT_ID = 'EXP'
    RESULT_ID = 'RST'
    METHOD_ID = 'MTD'
    REPORT_ID = 'RPT'
    TIMESTAMP = 'TS'
    VERSION = 'V'


from pydantic import BaseModel, model_validator
from labfreed.well_known_keys.unece.unece_units import unece_units


class Quantity(BaseModel):
    ''' Represents a quantity'''
    value: float|int
    unit: str
    significant_digits: int|None = None
    
    @model_validator(mode='after')
    def significat_digits_for_int(self):
        if isinstance(self.value, int):
            self.significant_digits = 0
        return self
        
    @property
    def float(self) -> float:
        ''' for clarity returns the value'''
        return self.value
    
    def __str__(self):
        unit_symbol = self.unit
        if self.unit == "dimensionless" or not self.unit:
            unit_symbol = ""
        if self.significant_digits is not None:
            val = f"{self.value:.{self.significant_digits}f}"
        else:
            val = str(self.value)
        return f"{val} {unit_symbol}"
      
    def __repr__(self):
        return f'Quantity: {self.__repr__()}'
    
    
    
    
def unece_unit_code_from_quantity(q:Quantity):
        by_name =   [ u['commonCode'] for u in unece_units() if u.get('name','') == q.unit] 
        by_symbol = [ u['commonCode'] for u in unece_units() if u.get('symbol','') == q.unit]
        code = list(set(by_name) | set(by_symbol))
        if len(code) != 1:
            raise ValueError(f'No UNECE unit code found for Quantity {str(q)}' ) 
        return code[0]
    
    



from datetime import date, datetime, time
import logging
import re



from pydantic import PrivateAttr, model_validator
from labfreed.labfreed_infrastructure import LabFREED_BaseModel, ValidationMsgLevel, _quote_texts
from abc import ABC, abstractmethod



''' Configure pdoc'''
v_segs  = ["NumericSegment", 
           "DateSegment", 
           "BoolSegment", 
           "AlphanumericSegment", 
           "TextSegment", 
           "BinarySegment"]
__all__ = ["TREX"] + v_segs + ["TableSegment"]  # noqa: F822

            
           

class Value(LabFREED_BaseModel, ABC):
    '''@private
    Helper to add validation for various types to ValueSegments and Tables
    '''
    value:str 
    
    def serialize(self):
        return self.value
        
        
        
class NumericValue(Value):
        
    @model_validator(mode='after')
    def _validate(self):
        value = self.value
        if not_allowed_chars := set(re.sub(r'[0-9\.\-E]', '', value)):
            self._add_validation_message(
                source=f"TREX numeric value {value}",
                level=ValidationMsgLevel.ERROR,
                msg=f"Characters {_quote_texts(not_allowed_chars)} are not allowed in quantity segment. Must be a number.",
                highlight_pattern = f'{value}',
                highlight_sub=not_allowed_chars
            )
        if not re.fullmatch(r'-?\d+(\.\d+)?(E-?\d+)?', value):
            self._add_validation_message(
                source=f"TREX numeric value {value}",
                level=ValidationMsgLevel.ERROR,
                msg=f"{value} cannot be converted to number",
                highlight_pattern = f'{value}'               
            )
        return self

class DateValue(Value):
    _date_time_dict:dict|None = PrivateAttr(default=None)
    
    @model_validator(mode='after')
    def _validate(self):
        pattern:str = r'((?P<year>\d{4})(?P<month>\d{2})(?P<day>\d{2}))?(T(?P<hour>\d{2})(?P<minute>\d{2})(?P<second>\d{2})?(\.(?P<millisecond>\d{3}))?)?'
        value=self.value
        if not re.fullmatch(pattern, value):
            self._add_validation_message(
                source=f"TREX date value {value}",
                level=ValidationMsgLevel.ERROR,
                msg=f'{value} is not in a valid format. Valid format for date: YYYYMMDD; Valid for time: THHMM, THHMMSS, THHMMSS.SSS; Datetime any combination of valid date and time',
                highlight_pattern = f'{value}'
            )
            return self
            
        matches = re.match(pattern, value)       
        d = matches.groupdict()  
        d = {k: int(v) for k,v in d.items() if v }
        if 'millisecond' in d.keys():
            ms = d.pop('millisecond')
            d.update({'microsecond': ms * 1000})
        try:
            if d.get('year'): # input is only a time
                datetime(**d)
            else:
                time(**d)
        except ValueError:
            self._add_validation_message(
                source=f"TREX date value {value}",
                level=ValidationMsgLevel.ERROR,
                msg=f'{value} is no valid date or time.',
                highlight_pattern = f'{value}'
            )
            
        self._date_time_dict = d
        return self
    

class BoolValue(Value):

    @model_validator(mode='after')
    def _validate(self):
        if self.value not in ['T', 'F']:
            self._add_validation_message(
                source=f"TREX boolean value {self.value}",
                level= ValidationMsgLevel.ERROR,
                msg=f'{self.value} is no valid boolean. Must be T or F',
                highlight_pattern = f'{self.value}',
                highlight_sub=[c for c in self.value]
            )
        return self
                  
                    
class AlphanumericValue(Value):
        
    @model_validator(mode='after')
    def _validate(self):
        if re.match(r'[a-z]', self.value):
            self._add_validation_message(
                    source=f"TREX value {self.value}",
                    level= ValidationMsgLevel.ERROR,
                    msg="Lower case characters are not allowed.",
                    highlight_pattern = self.value
            )
            
        if not_allowed_chars := set(re.sub(r'[A-Z0-9\.-]', '', self.value)):
            self._add_validation_message(
                    source=f"TREX value {self.value}",
                    level= ValidationMsgLevel.ERROR,
                    msg=f"Characters {_quote_texts(not_allowed_chars)} are not allowed in alphanumeric segment",
                    highlight_pattern = self.value,
                    highlight_sub=not_allowed_chars
            )
        return self
             
    
class TextValue(Value):
        
    @model_validator(mode='after')
    def _validate(self):
        if not_allowed_chars := set(re.sub(r'[A-Z0-9]', '', self.value)):
            self._add_validation_message(
                    source=f"TREX value {self.value}",
                    level= ValidationMsgLevel.ERROR,
                    msg=f"Characters {_quote_texts(not_allowed_chars)} are not allowed in text segment. Base36 encoding only allows A-Z0-9",
                    highlight_pattern = self.value,
                    highlight_sub=not_allowed_chars
            )
        return self
                
    
class BinaryValue(Value):
        
    @model_validator(mode='after')
    def _validate(self):
        if not_allowed_chars := set(re.sub(r'[A-Z0-9]', '', self.value)):
           self._add_validation_message(
                    source=f"TREX value {self.value}",
                    level= ValidationMsgLevel.ERROR,
                    msg=f"Characters {_quote_texts(not_allowed_chars)} are not allowed in text segment. Base36 encoding only allows A-Z0-9",
                    highlight_pattern = self.value,
                    highlight_sub=not_allowed_chars
            )
        return self
                 
    
class ErrorValue(Value):
    
    @model_validator(mode='after')
    def _validate(self):
        if not_allowed_chars := set(re.sub(r'[A-Z0-9\.-]', '', self.value)):
            self._add_validation_message(
                    source=f"TREX value {self.value}",
                    level= ValidationMsgLevel.ERROR,
                    msg=f"Characters {_quote_texts(not_allowed_chars)} are not allowed in error segment",
                    highlight_pattern = self.value,
                    highlight_sub=not_allowed_chars
            )
        return self


class TREX_Segment(LabFREED_BaseModel, ABC):
    '''@private
    Abstract class representing a TREX_Segment
    '''
    key: str
    
    @abstractmethod
    def serialize(self):
        raise NotImplementedError("Subclasses must implement 'serialize_as_trex()' method")
    
    
    @model_validator(mode='after')
    def _validate_key(self):
        if not_allowed_chars := set(re.sub(r'[A-Z0-9\.-]', '', self.key)):
            self._add_validation_message(
                source=f"TREX segment key {self.key}",
                level=ValidationMsgLevel.ERROR,
                msg=f"Segment key contains invalid characters: {_quote_texts(not_allowed_chars)}",
                highlight_pattern = f'{self.key}$',
                highlight_sub=not_allowed_chars
            )
        return self
    

        



    





import logging
from typing import Self
import yaml
from requests import get



from labfreed.pac_cat.pac_cat import PAC_CAT
from labfreed.pac_id.pac_id import PAC_ID
from labfreed.pac_id_resolver.services import ServiceGroup
from labfreed.pac_id_resolver.cit_v1 import CIT_v1
from labfreed.pac_id_resolver.cit_v2 import CIT_v2



''' Configure pdoc'''
__all__ = ["PAC_ID_Resolver"]

def load_cit(path):
    with open(path, 'r') as f:
        s = f.read()
        return _cit_from_str(s)

    
def _cit_from_str(s:str, issuer:str='') -> CIT_v1|CIT_v2:
    try:
        cit_yml= yaml.safe_load(s)
        cit2 = CIT_v2.from_yaml(cit_yml)
    except Exception:
        cit2 = None
    
    try:
        cit1 = CIT_v1.from_csv(s, issuer)
    except Exception:
        cit1 = None
    
    cit = cit2 or cit1 or None
    return cit

def _get_issuer_cit(issuer:str):
    '''Gets the issuer's cit.'''
    url = 'HTTPS://PAC.' + issuer + '/coupling-information-table'
    try:
        r = get(url, timeout=2)
        if r.status_code < 400:
            cit_str = r.text
        else: 
            logging.error(f"Could not get CIT form {issuer}")
            cit_str  = None
    except Exception:
        logging.error(f"Could not get CIT form {issuer}")
        cit_str  = None
    cit = _cit_from_str(cit_str, issuer=issuer)
    return cit
    


class PAC_ID_Resolver():
    def __init__(self, cits:list[CIT_v2|CIT_v1]=None) -> Self:
        '''Initialize the resolver with coupling information tables'''
        if not cits:
            cits = []
        self._cits = cits
            
        
    def resolve(self, pac_id:PAC_ID|str) -> list[ServiceGroup]:
        '''Resolve a PAC-ID'''
        if isinstance(pac_id, str):
            pac_id = PAC_CAT.from_url(pac_id)
                
        if issuer_cit := _get_issuer_cit(pac_id.issuer):
            self._cits.append(issuer_cit)
         
        matches = [cit.evaluate_pac_id(pac_id) for cit in self._cits]
        return matches
            
    
    
if __name__ == '__main__':
    r = PAC_ID_Resolver()
    r.resolve()



''' 
Python implementation of LabFREED building blocks
'''

__version__ = "0.2.5a6"

from labfreed.pac_id import *  # noqa: F403
from labfreed.pac_cat import * # noqa: F403
from labfreed.pac_id_resolver import * # noqa: F403
from labfreed.trex import * # noqa: F403
from labfreed.labfreed_infrastructure import *  # noqa: F403



import re
import string

from pydantic import field_validator, RootModel

class base36(RootModel[str]):
    @field_validator('root')
    @classmethod
    def validate_format(cls, v: str) -> str:
        if not re.fullmatch(r'[A-Z0-9]*', v):
            raise ValueError("Value must only contain uppercase letters and digits (A-Z, 0-9)")
        return v
    

def to_base36(s:str) -> base36:
    """Takes a string, encodes it in UTF-8 and then as base36 string."""
    utf8_encoded = s.encode('utf-8')
    num = int.from_bytes(utf8_encoded, byteorder='big', signed=False)
    
    # note: this cannot be arbitrarily chosen. The choice here corresponds to what pythons int(s:str, base:int=10) function used.
    base36_chars = _alphabet(base=36)
    if num == 0:
        return base36_chars[0]
    base_36 = []
    _num = num
    while _num:
        _num, i = divmod(_num, 36)
        base_36.append(base36_chars[i])
    b36_str = ''.join(reversed(base_36))
    b36_str = base36(b36_str)
    return b36_str


def from_base36(s36:base36) -> str:
    """inverse of to_base36"""
    # this built in function interprets each character as number in a base represented by the standartd alphabet [0-9 (A-Z|a-z)][0:base] it is case INsensitive.
    num = int(s36, 36)
    num_bytes = (num.bit_length() + 7) // 8
    _bytes = num.to_bytes(num_bytes, byteorder='big')
    s = _bytes.decode('utf-8')
    return s


def _alphabet(base):
    """ returns an alphabet, which corresponds to what pythons int(s:str, base:int=10) function used.
    """
    if base < 2 or base > 36:
        ValueError('base can only be between 2 and 36')    
    alphabet = (string.digits + string.ascii_uppercase)[0:base]
    return alphabet


if __name__ == "__main__":
    ss = ["A",
      "B-500 B",
      "B-500 Ba",
      "B-500 Bal",
      "B-500 Bala",
      "B-500 Balanc",
      "B-500 Balance",
      "B-500 D",
      "Mini Spray Dryer S-300", 
      "w3ApashAt!!£NAGDSAF*ç%&/()",
      "HELLOWORLD", 
      "Helloworld",
      "$£äö!'?^{]<@#¦&¬|¢)&§°😀你好🌍🏯😇🎵🔥你👻🐉😀你好🌍🏯😇🎵🔥你👻🐉😀你好🌍🏯😇🎵🔥你👻🐉😀你好🌍🏯😇🎵🔥你👻🐉😀你好🌍🏯😇🎵🔥你👻🐉😀你好🌍🏯😇🎵🔥你👻🐉😀你好🌍🏯😇🎵🔥你👻🐉",
      "往跟住！師立甲錯什正再圓身升因月室",
      "Balance BAL500 @☣️Lab",
      "BAL500 @☣️Lab",
      "BAL-CLEAN",
      "Smørrebrød µ-Nutrients",
      "Demo Result from R-300",
      "Rotavapor R-300",
      "Rotavapor R-250",
      "Rotavapor R-220",
      "SyncorePlus"
      ]
    for s in ss:
        s36 = to_base36(s)
        s_back = from_base36(s36)
        identical = (s == s_back)
        print(f'{s} >> {s36} >> {s_back}: match:{identical}')



import re
from typing_extensions import Self
from pydantic import Field, conlist, model_validator

from labfreed.labfreed_infrastructure import LabFREED_BaseModel, ValidationMsgLevel
from labfreed.pac_id.id_segment import IDSegment
from labfreed.pac_id.extension import Extension


from typing import TYPE_CHECKING
if TYPE_CHECKING:
    pass


_domain_name_pattern = r"(?!-)([A-Za-z0-9-]{1,63}(?<!-)\.)+[A-Za-z]{2,63}"

class PAC_ID(LabFREED_BaseModel):
    '''Represents a PAC-ID. 
    Refer to the [specification](https://github.com/ApiniLabs/PAC-ID?tab=readme-ov-file#specification) for details.
    '''
    issuer:str 
    '''The issuer of the PAC-ID.'''
    identifier: conlist(IDSegment) = Field(..., default_factory=list) # type: ignore # exclude=True prevents this from being serialized by Pydantic
    '''The identifier of the PAC-ID is a series of IDSegments'''
    
    extensions: list[Extension] = Field(default_factory=list)
    
    
    def get_extension_of_type(self, type:str) -> list[Extension]:
        '''Get all extensions of a certain type.'''
        return [e for e in self.extensions if e.type == type]
    
    
    def get_extension(self, name:str) -> Extension|None:
        '''Get extension of certain name'''
        out = [e for e in self.extensions if e.name == name]
        if not out:
            return None
        return out[0]
    
    @classmethod
    def from_url(cls, url, *, extension_interpreters='default', 
                 try_pac_cat=True,
                 suppress_validation_errors=False) -> Self:
        from labfreed.pac_id.url_parser import PAC_Parser
        return PAC_Parser.from_url(url, try_pac_cat=try_pac_cat, suppress_validation_errors=suppress_validation_errors, extension_interpreters=extension_interpreters)
    
    def to_url(self, use_short_notation=False, uppercase_only=False) -> str:
        from labfreed.pac_id.url_serializer import PACID_Serializer
        return PACID_Serializer.to_url(self, use_short_notation=use_short_notation, uppercase_only=uppercase_only)
    
    def to_json(self, indent=None) -> str:
        if not indent:
            return self.model_dump_json()
        else:
            return self.model_dump_json(indent=indent)
        
    def to_dict(self) -> dict:
        return self.model_dump()
    
    def __str__(self):
        return self.to_url()
          
         
    @model_validator(mode='after')
    def _check_at_least_one_segment(self) -> Self:
        if not len(self.identifier) >= 1:
            self._add_validation_message(
                        source="identifier",
                        level = ValidationMsgLevel.ERROR,
                        msg='Identifier must contain et least one segment.'
                    )
        return self
    
        
    @model_validator(mode='after')
    def _check_length(self) -> Self:
        length = 0
        for s in self.identifier:
            s:IDSegment = s
            if s.key:
                length += len(s.key)
                length += 1 # for ":"
            length += len(s.value)
        length += len(self.identifier) - 1 # account for "/" separating the segments
        
        if length > 256:
            self._add_validation_message(
                        source="identifier",
                        level = ValidationMsgLevel.ERROR,
                        msg=f'Identifier is {length} characters long, Identifier must not exceed 256 characters.'
                    )
        return self
       
    
    @model_validator(mode="after")
    def _validate_issuer(self):
        if not re.fullmatch(_domain_name_pattern, self.issuer):
            self._add_validation_message(
                    source="PAC-ID",
                    level = ValidationMsgLevel.ERROR,
                    highlight_pattern=self.issuer,
                    msg="Issuer must be a valid domain name. "
                )
         
        # recommendation that A-Z, 0-9, -, and . should be used
        if not_recommended_chars := set(re.sub(r'[A-Z0-9\.-]', '', self.issuer)):
            self._add_validation_message(
                    source="PAC-ID",
                    level = ValidationMsgLevel.RECOMMENDATION,
                    highlight_pattern=self.issuer,
                    highlight_sub=not_recommended_chars,
                    msg=f"Characters {' '.join(not_recommended_chars)} should not be used. Issuer SHOULD contain only the characters A-Z, 0-9, -, and . "
                )
        return self
    
    
    @model_validator(mode='after')
    def _check_identifier_segment_keys_are_unique(self) -> Self:
        keys = [s.key for s in self.identifier if s.key]
        duplicate_keys = [k for k in set(keys) if keys.count(k) > 1]
        if duplicate_keys:
            for k in duplicate_keys:
                self._add_validation_message(
                    source=f"identifier {k}",
                    level = ValidationMsgLevel.RECOMMENDATION,
                    msg=f"Duplicate segment key {k}. This will probably lead to undefined behaviour",
                    highlight_pattern = k
                )
        return self
    
    

    
    

    
    





from typing import Literal, Self

from pydantic import computed_field 
from labfreed.labfreed_infrastructure import LabFREED_BaseModel
from labfreed.pac_id.extension import ExtensionBase
from labfreed.trex.trex import TREX


class TREX_Extension(ExtensionBase, LabFREED_BaseModel):
    name:str
    type:Literal['TREX'] = 'TREX'
    trex:TREX
           
    @computed_field
    @property
    def data(self)->str:
        trex_str = self.trex.serialize()
        return trex_str
        
    @staticmethod
    def from_extension(ext:ExtensionBase) -> Self:
        return TREX_Extension.create(name=ext.name,
                                    type=ext.type,
                                    data=ext.data)
    
    @staticmethod
    def create(*, name, data, type='TREX'):
        trex_extension = TREX_Extension(name= name, trex = TREX.deserialize(data))
        return trex_extension
    
   


from .trex import TREX
from .value_segments import NumericSegment, DateSegment, BoolSegment, AlphanumericSegment, TextSegment, ErrorSegment
from .table_segment import TableSegment, ColumnHeader, TableRow

__all__ = [
    "TREX",
    "NumericSegment",
    "DateSegment",
    "BoolSegment",
    "AlphanumericSegment",
    "TextSegment",
    "ErrorSegment",
    "TableSegment",
    "ColumnHeader",
    "TableRow"
]



               
from typing import Any
from pydantic import PrivateAttr, computed_field, model_validator
from labfreed.labfreed_infrastructure import LabFREED_BaseModel, ValidationMsgLevel
from labfreed.pac_id.id_segment import IDSegment


class Category(LabFREED_BaseModel):
    '''
    Represents a category. \n
    This is the base class for categories. If possible a more specific category should be used.
    '''    
    key:str
    '''The category key, e.g. "-MD"'''
    _segments: list[IDSegment] = PrivateAttr(default_factory=list)
    
        
    @computed_field
    @property
    def segments(self) -> list[IDSegment]:
        return self._segments
    
    def __init__(self, **data: Any):
        '''@private'''
        # Pop the user-provided value for computed segments
        input_segments = data.pop("segments", None)
        super().__init__(**data)
        self._segments = input_segments
    
    @model_validator(mode='after')
    def _warn_unusual_category_key(self):
        ''' this base class is instantiated only if the key is not a known category key'''
        if type(self) is Category:
            self._add_validation_message(
                        source=f"Category {self.key}",
                        level = ValidationMsgLevel.RECOMMENDATION,
                        msg=f'Category key {self.key} is not a well known key. It is recommended to use well known keys only',
                        highlight_pattern = f"{self.key}"
            )
        return self
    
    
    def __str__(self):
        s = '\n'.join( [f"{field_name} \t ({field_info.alias or ''}): \t {getattr(self, field_name)}" for  field_name, field_info in self.model_fields.items() if getattr(self, field_name)]) 
        return s 









import logging
from typing import Literal, Self
from pydantic import computed_field
from labfreed.labfreed_infrastructure import LabFREED_BaseModel
from labfreed.pac_id.extension import ExtensionBase
from labfreed.utilities.base36 import from_base36, to_base36


class DisplayNameExtension(ExtensionBase, LabFREED_BaseModel):
    name:Literal['N'] = 'N'
    type:Literal['N'] = 'N'
    display_name: str       
    
    @computed_field
    @property
    def data(self)->str:
        # return '/'.join([to_base36(dn) for dn in self.display_name])
        return to_base36(self.display_name).root
    
    @staticmethod
    def from_extension(ext:ExtensionBase) -> Self:
        return DisplayNameExtension.create(name=ext.name,
                                  type=ext.type,
                                  data=ext.data)
    
    @staticmethod
    def create(*, name, type, data):
        if name != 'N':
            logging.warning(f'Name {name} was given, but this extension should only be used with name "N". Will ignore input')
            
        if type != 'N':
            logging.warning(f'Type {name} was given, but this extension should only be used with type "N". Will try to parse data as display names')
        
        display_name = from_base36(data)
         
        return DisplayNameExtension(display_name=display_name)
    
    def __str__(self):
        return 'Display name: '+ self.display_name




from enum import Enum

class GS1ApplicationIdentifier(Enum):
    SSCC = "00"
    GTIN = "01"
    CONTENT = "02"
    MTO_GTIN = "03"
    BATCH_LOT = "10"
    PROD_DATE = "11"
    DUE_DATE = "12"
    PACK_DATE = "13"
    BEST_BEFORE_OR_BEST_BY = "15"
    SELL_BY = "16"
    USE_BY_OR_EXPIRY = "17"
    VARIANT = "20"
    SERIAL = "21"
    CPV = "22"
    VAR_COUNT = "30"
    COUNT = "37"
    INTERNAL = "90"
    TPX = "235"
    ADDITIONAL_ID = "240"
    CUST_PART_NO = "241"
    MTO_VARIANT = "242"
    PCN = "243"
    SECONDARY_SERIAL = "250"
    REF_TO_SOURCE = "251"
    GDTI = "253"
    GLN_EXTENSION_COMPONENT = "254"
    GCN = "255"
    ORDER_NUMBER = "400"
    GINC = "401"
    GSIN = "402"
    ROUTE = "403"
    SHIP_TO_LOC = "410"
    BILL_TO = "411"
    PURCHASE_FROM = "412"
    SHIP_FOR_LOC = "413"
    LOC_NO = "414"
    PAY_TO = "415"
    PROD_SERV_LOC = "416"
    PARTY = "417"
    SHIP_TO_POST = "420"
    ORIGIN = "422"
    COUNTRY___INITIAL_PROCESS = "423"
    COUNTRY___PROCESS = "424"
    COUNTRY___DISASSEMBLY = "425"
    COUNTRY___FULL_PROCESS = "426"
    ORIGIN_SUBDIVISION = "427"
    NHRN_PZN = "710"
    NHRN_CIP = "711"
    NHRN_CN = "712"
    NHRN_DRN = "713"
    NHRN_AIM = "714"
    NHRN_NDC = "715"
    NHRN_AIC = "716"
  



from .pac_cat import PAC_CAT
from .category_base import Category
from .predefined_categories import (
    Material_Device, Material_Substance, Material_Consumable, Material_Misc, Data_Method, Data_Result, Data_Progress, 
    Data_Calibration, Data_Abstract, category_key_to_class_map
    )

__all__ = [
    "PAC_CAT",
    "Category",
    "Material_Device",
    "Material_Substance",
    "Material_Consumable",
    "Material_Misc",
    "Data_Method",
    "Data_Result",
    "Data_Progress",
    "Data_Calibration",
    "Data_Abstract"
]